// SigWorld - Chris's version
//
// By Chris (chrish@qnx.com)
// April 8, 1998
//
// This is a simple demo to show how you might want to deal with POSIX
// signals in a “real” BeOS application.
//
// This version really is signal-safe, and uses a "proxy" technique; a
// thread is spawned before we start handling signals… it's this thread's
// job to send us a message when we get a signal in the signal handler.

// ----------------------------------------------------------------------
// All the usual BeOS headers.
#include <app/Application.h>
#include <app/Message.h>
#include <interface/Window.h>
#include <interface/View.h>
#include <kernel/OS.h>

// ----------------------------------------------------------------------
// POSIX 1003.1 headers; everything you need for dealing with signals,
// errno, and stdio.
#include <signal.h>
#include <errno.h>
#include <stdio.h>

// ----------------------------------------------------------------------
// The Proxy
//
// The proxy thread waits around, blocking on a semaphore; the signal
// handler will release the semaphore, which lets the proxy send its
// message to the application.
//
// This version re-arms itself after sending the message; the application
// might be ignoring B_QUIT_REQUESTED for now, so we should be prepared to
// handle the signal again.
//
// For a one-shot proxy, you'd just remove the “while” loop.

int32 proxy_thread( void *data );
sem_id proxy_sem;

int32 proxy_thread( void *data )
{
#pragma unused (data)

	while( true ) {
		status_t retval = acquire_sem( proxy_sem );
		if( retval == B_NO_ERROR ) {
			// Then we were released and should send a message.
			printf( "Proxy thread triggered… sending B_QUIT_REQUESTED.\n" );
			be_app->PostMessage( B_QUIT_REQUESTED );
		} else {
			// Then the semaphore is gone, and we should exit.
			printf( "Proxy semaphore is gone… proxy thread exitting.\n" );
			return 0;
		}
	}

	// You'll never get here; the proxy should bail out when the
	// semaphore is deleted.
	return 0;
}

// ----------------------------------------------------------------------
// Our illustrious signal handler
//
// This signal handler just unblocks the proxy.

void handle_signals( int sig );

void handle_signals( int sig )
{
	if( sig == SIGINT ) {
		// This is the signal we want to handle, so we trigger the proxy.
		status_t retval = release_sem( proxy_sem );
	}

	return;
}

// ----------------------------------------------------------------------
// Our view object.
class SigView : public BView {
public:
	SigView( BRect frame, const char *name, uint32 resizing, uint32 flags )
		: BView( frame, name, resizing, flags ) { ; }
	virtual ~SigView( void ) { ; }
	
	virtual void Draw( BRect update );
};

void SigView::Draw( BRect update )
{
#pragma unused (update)
	MovePenTo( 10.0, 25.0 );
	DrawString( "Do a “kill -INT chrish” in a Terminal!" );
}

// ----------------------------------------------------------------------
// Our application object.
class SigApplication : public BApplication {
public:
	SigApplication( void ) : BApplication( "application/x-vnd.cjh-sigworld" ) {
		_triggered = false;

		_wind = new BWindow( BRect( 100, 100, 600, 200 ), "SigWorld - Chris’s version",
						B_TITLED_WINDOW, 
						B_NOT_RESIZABLE | B_NOT_ZOOMABLE | B_NOT_MINIMIZABLE );
		if( _wind ) {
			BView *view = new SigView( _wind->Bounds(), "Sig view",
								B_FOLLOW_ALL_SIDES, B_WILL_DRAW );
			if( view ) {
				_wind->AddChild( view );
			}

			_wind->Show();
		}
	}

	virtual ~SigApplication( void ) { ; }

	virtual bool QuitRequested( void );

private:
	BWindow *_wind;
	bool _triggered;
};

bool SigApplication::QuitRequested( void )
{
	// The first time we get QuitRequested() we'll ignore it.
	if( _triggered ) {
		printf( "SigApplication got a QuitRequested… we're outta here.\n" );

		return true;
	}

	printf( "SigApplication got a QuitRequested… ignoring it.\n" );
	_triggered = true;

	return false;
}

// ----------------------------------------------------------------------
// Mainline
int main( int argc, char **argv )
{
#pragma unused (argc,argv)
	// We need to set up our proxy semaphore and thread before we even
	// thing about handling any signals.
	proxy_sem = create_sem( 0, "proxy semaphore" );
	if( proxy_sem < B_NO_ERROR ) {
		printf( "Can't create the proxy semaphore; bailing out.\n" );
		return -1;
	}
	
	thread_id proxy_tid = spawn_thread( proxy_thread, "proxy thread",
								B_NORMAL_PRIORITY, NULL );
	if( proxy_tid < B_NO_ERROR ) {
		printf( "Can't create the proxy thread; bailing out.\n" );
		status_t retval = delete_sem( proxy_sem );
		return -1;
	}

	// Now launch the proxy thread, which will immediately block.
	status_t retval = resume_thread( proxy_tid );
	if( retval != B_NO_ERROR ) {
		printf( "Can't start the proxy thread; bailing out.\n" );
		status_t retval = delete_sem( proxy_sem );
		return -1;
	}

	// Set up our signal handler.  We'll only handle SIGINT here, just
	// in case things don't work out... you can still send a SIGKILL to
	// take out the app.
	sigset_t set;
	(void)sigemptyset( &set );
	(void)sigaddset( &set, SIGINT );	// The signal we want to catch.

	struct sigaction handler;
	handler.sa_handler = handle_signals;
	(void)sigfillset( &handler.sa_mask );	// ignore all signals during handle_signals()
	handler.sa_flags = 0;

	int sigret = sigaction( SIGINT, &handler, (struct sigaction *)NULL );
	if( sigret != EOK ) {
		switch( sigret ) {
		case EINVAL:
			printf( "attempted to set up handler for invalid signal\n" );
			break;
		default:
			printf( "unknown error from sigaction()\n" );
			break;
		}

		// This will give you a warning; "Illegal jump past initializer".
		// That's the compiler reminding you that app won't be a valid
		// variable at the jump target.
		//
		// Of course, we've been careful to make sure we don't try to
		// use app after exit_cleanly, so we'll ignore the warning.
		goto exit_cleanly;
	}

	// Now create our app and let it run.
	SigApplication app;
	app.Run();

	// Clean up before exitting; by deleting the semaphore, we're allowing
	// our proxy thread to exit cleanly.
exit_cleanly:
	retval = delete_sem( proxy_sem );
	retval = wait_for_thread( proxy_tid, NULL );

	// Now we exit…
	return 0;
}
